# Copyright 2014 Alex Orange
# 
# This file is part of LDB Algebra.
# 
# LDB Algebra is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# LDB Algebra is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with LDB Algebra.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

import operator

def freezedict(dict_):
    return tuple(sorted(dict_.items()))

class MethodCache(object):
    def __init__(self):
        self.order = []
        self.cache = {}
        self.max_len = 5

    def __contains__(self, key):
        return key in self.order

    def get_result(self, key):
        # TODO: Test performance for test and test and grab vs grab and handle
        # exception
        if key in self.order:
            return self.cache[key]

    def store_result(self, key, val):
        if key in self.order:
            self.order.remove(key)
        elif len(self.order) >= self.max_len:
            to_kill = self.order[0]
            del self.cache[to_kill]
            del self.order[0]

        self.cache[key] = val
        self.order.append(key)


def bind(expression, binding):
    if hasattr(expression, 'bind'):
        return expression.bind(**binding)
    else:
        return expression


def differentiate(expression, variable):
    if hasattr(expression, 'differentiate'):
        return expression.differentiate(variable)
    else:
        return 0


class Expression(object):
    def __init__(self):
        pass

    def __mul__(self, other):
        if other == 0:
            return 0
        elif other == 1:
            return self
        return Product(self, other)

    def __rmul__(self, other):
        if other == 0:
            return 0
        elif other == 1:
            return self
        return Product(other, self)

    def __add__(self, other):
        if other == 0:
            return self
        return Sum(self, other)

    def __radd__(self, other):
        if other == 0:
            return self
        return Sum(other, self)

    def __sub__(self, other):
        if self == other:
            return 0
        if other == 0:
            return self
        return Sum(self, -1*other)

    def __rsub__(self, other):
        if self == other:
            return 0
        if other == 0:
            return -self
        return Sum(other, -1*self)

    def __truediv__(self, other):
        if other == 1:
            return self
        return Quotient(self, other)

    def __rtruediv__(self, other):
        if other == 0:
            return 0
        return Quotient(other, self)

    def __div__(self, other):
        if other == 1:
            return self
        return Quotient(self, other)

    def __rdiv__(self, other):
        if other == 0:
            return 0
        return Quotient(other, self)

    def __neg__(self):
        return -1*self

    def __pow__(self, other):
        return Power(self, other)


class Variable(Expression):
    def __init__(self, var_name):
        self.var_name = var_name

    def bind(self, **binding):
        if self.var_name in binding:
            return binding[self.var_name]
        else:
            return self

    def differentiate(self, variable):
        if self.var_name == variable:
            return 1
        else:
            return 0

    def __repr__(self):
        return self.var_name


class CachedExpression(Expression):
    def __init__(self):
        self._binding_cache = MethodCache()
        self._differentiate_cache = MethodCache()

    def bind(self, **binding):
        hash_binding = freezedict(binding)
        if hash_binding in self._binding_cache:
            return self._binding_cache.get_result(hash_binding)
        else:
            val = self._bind(binding)
            self._binding_cache.store_result(hash_binding, val)
            return val

    def differentiate(self, variable):
        if variable in self._differentiate_cache:
            return self._differentiate_cache.get_result(variable)
        else:
            val = self._differentiate(variable)
            self._differentiate_cache.store_result(variable, val)
            return val


class Product(CachedExpression):
    def __init__(self, a, b):
        super(Product, self).__init__()
        self.a = a
        self.b = b

    def _bind(self, binding):
        return bind(self.a, binding) * bind(self.b, binding)

    def _differentiate(self, variable):
        da = differentiate(self.a, variable)
        db = differentiate(self.b, variable)
        return self.b * da + self.a * db

    def __repr__(self):
        return "(%s * %s)"%(repr(self.a), repr(self.b))


class Quotient(CachedExpression):
    def __init__(self, a, b):
        super(Quotient, self).__init__()
        self.a = a
        self.b = b

    def _bind(self, binding):
        return operator.truediv(bind(self.a, binding), bind(self.b, binding))

    def _differentiate(self, variable):
        da = differentiate(self.a, variable)
        db = differentiate(self.b, variable)
        return (self.b * da - self.a * db) / (self.b*self.b)

    def __repr__(self):
        return "(%s / %s)"%(repr(self.a), repr(self.b))


class Sum(CachedExpression):
    def __init__(self, a, b):
        super(Sum, self).__init__()
        self.a = a
        self.b = b

    def _bind(self, binding):
        return bind(self.a, binding) + bind(self.b, binding)

    def _differentiate(self, variable):
        da = differentiate(self.a, variable)
        db = differentiate(self.b, variable)
        return da + db

    def __repr__(self):
        return "(%s + %s)"%(repr(self.a), repr(self.b))


class Power(CachedExpression):
    def __init__(self, a, b):
        super(Power, self).__init__()
        self.a = a
        self.b = b

    def _bind(self, binding):
        return bind(self.a, binding) ** bind(self.b, binding)

    def _differentiate(self, variable):
        import ldb.algebra.math

        da = differentiate(self.a, variable)
        db = differentiate(self.b, variable)
        return (da * self.b * (self.a ** (self.b - 1)) +
                db * ldb.algebra.math.log(self.a) * (self.a ** self.b))

    def __repr__(self):
        return "(%s ** %s)"%(self.a, self.b)



# Copyright 2014 Alex Orange
# 
# This file is part of LDB Algebra.
# 
# LDB Algebra is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# LDB Algebra is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with LDB Algebra.  If not, see <http://www.gnu.org/licenses/>.

from expression import *

def distribute(expression):
    """
    >>> from ldb.algebra.expression import Variable
    >>> a = Variable('a')
    >>> b = Variable('b')
    >>> distribute(a+b)
    [([a], []), ([b], [])]
    >>> distribute(a*b)
    [([a, b], [])]
    >>> distribute(a/b)
    [([a], [b])]
    """
    # TODO: Add more thorough tests
    # List of (numerator product terms, denominator product terms) empty lists
    # in tuple interpreted as 1.0
    sum_ = []
    if isinstance(expression, Sum):
        sum_ += distribute(expression.a)
        sum_ += distribute(expression.b)
    elif isinstance(expression, Product):
        a = distribute(expression.a)
        b = distribute(expression.b)
        for a_term in a:
            for b_term in b:
                sum_.append((a_term[0]+b_term[0], a_term[1]+b_term[1]))
    elif isinstance(expression, Quotient):
        a = distribute(expression.a)
        sum_ += [(_[0], _[1]+[expression.b]) for _ in a]
    else:
        sum_.append(([expression], []))

    return sum_

def undistribute_product(product_):
    if len(product_) == 0:
        return 1.0

    result = product_[0]

    for term in product_[1:]:
        result *= term

    return result

def undistribute_quotient(numerator, denominator):
    num_expr = undistribute_product(numerator)

    if len(denominator) == 0:
        return num_expr
    else:
        return num_expr / undistribute_product(denominator)

def undistribute(sum_):
    """
    >>> from ldb.algebra.expression import Variable
    >>> a = Variable('a')
    >>> b = Variable('b')
    >>> undistribute([([a], []), ([b], [])])
    (a + b)
    >>> undistribute([([a, b], [])])
    (a * b)
    >>> undistribute([([a], [b])])
    (a / b)
    """
    # TODO: Add more thorough tests
    if len(sum_) == 0:
        return 0

    result = undistribute_quotient(*sum_[0])

    for term in sum_[1:]:
        result += undistribute_quotient(*term)

    return result

def divide(a, b):
    """
    divide(a, b) - returns (c, d) for expression b*c+d, b should be something
    that distributes to [([b], [])]
    >>> from ldb.algebra.expression import Variable
    >>> a = Variable('a')
    >>> b = Variable('b')
    >>> c = Variable('c')
    >>> divide(a*b+c, a)
    (b, c)
    """
    # TODO: Add more thorough tests
    c_terms = []
    d_terms = []

    for term in distribute(a):
        if b in term[0]:
            term[0].remove(b)
            c_terms.append(term)
        else:
            d_terms.append(term)

    return (undistribute(c_terms), undistribute(d_terms))



# Copyright 2014 Alex Orange
# 
# This file is part of LDB Algebra.
# 
# LDB Algebra is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# LDB Algebra is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with LDB Algebra.  If not, see <http://www.gnu.org/licenses/>.

from expression import CachedExpression

class Function(CachedExpression):
    def __init__(self, fname, derivatives=None):
        super(Function, self).__init__()
        self.fname = fname
        self.derivatives = derivatives
        if self.derivatives is None:
            self.derivatives = []
        self.derivatives.sort()

    def _bind(self, binding):
        repr_self = repr(self)
        if repr_self in binding:
            return binding[repr_self]
        else:
            return self

    def _differentiate(self, variable):
        return Function(self.fname, self.derivatives+[variable])

    def __eq__(self, other):
        return self.fname == other.fname and \
                self.derivatives == other.derivatives

    def __repr__(self):
        if len(self.derivatives) == 0:
            return self.fname

        dvars = list(set(self.derivatives))
        dvars.sort()
        ddata = [(self.derivatives.count(dvar), dvar) for dvar in dvars]
        dstr = ''.join(['d%d%s'%(_[0], _[1]) if _[0] > 1 else 'd%s'%(_[1])
                        for _ in ddata])
        dfstr = 'd%d%s'%(len(self.derivatives), self.fname) \
                    if len(self.derivatives) > 1 else 'd%s'%(self.fname)
        return '%s_%s'%(dfstr, dstr)



# Copyright 2015 Alex Orange
# 
# This file is part of LDB Algebra.
# 
# LDB Algebra is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# LDB Algebra is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with LDB Algebra.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

import math
from ldb.algebra.expression import (CachedExpression, Expression, bind,
                                    differentiate)


def wrap_function(function_class, function):
    function_class.function = function
    return lambda x: function_class(x) if isinstance(x, Expression) else function(x)


class WrappedFunction(object):
    def __init__(self, function):
        self.function = function

    def __call__(self, expression):
        if isinstance(expression, Expression):
            return self.function_class(expression)
        else:
            return self.function(expression)


class MathFunction(CachedExpression):
    def __init__(self, a):
        super(MathFunction, self).__init__()
        self.a = a

    def _bind(self, binding):
        bound = bind(self.a, binding)
        if isinstance(bound, Expression):
            return self.__class__(bound)
        else:
            return self.function(bound)

    def _differentiate(self, variable):
        da = differentiate(self.a, variable)
        return da * self.derivative


class SquareRoot(MathFunction):
    def __init__(self, a):
        super(SquareRoot, self).__init__(a)

    @property
    def derivative(self):
        return 0.5 / self

    def __repr__(self):
        return "sqrt(%s)"%(repr(self.a))


class Logarithm(MathFunction):
    def __init__(self, a):
        super(Logarithm, self).__init__(a)

    @property
    def derivative(self):
        return 1 / self.a

    def __repr__(self):
        return "log(%s)"%(repr(self.a))


class ACos(CachedExpression):
    def __init__(self, a):
        super(ACos, self).__init__()
        self.a = a

    def _bind(self, binding):
        return acos(bind(self.a, binding))

    def _differentiate(self, variable):
        da = differentiate(self.a, variable)
        return -da / sqrt(1.0 - self.a*self.a)

    def __repr__(self):
        return "acos(%s)"%(repr(self.a))


class ACosh(CachedExpression):
    def __init__(self, a):
        super(ACosh, self).__init__()
        self.a = a

    def _bind(self, binding):
        return acosh(bind(self.a, binding))

    def _differentiate(self, variable):
        da = differentiate(self.a, variable)
        return da / sqrt(self.a*self.a - 1.0)

    def __repr__(self):
        return "acosh(%s)"%(repr(self.a))


class ASin(CachedExpression):
    def __init__(self, a):
        super(ASin, self).__init__()
        self.a = a

    def _bind(self, binding):
        return asin(bind(self.a, binding))

    def _differentiate(self, variable):
        da = differentiate(self.a, variable)
        return da / sqrt(1.0 - self.a*self.a)

    def __repr__(self):
        return "asin(%s)"%(repr(self.a))


class ASinh(CachedExpression):
    def __init__(self, a):
        super(ASinh, self).__init__()
        self.a = a

    def _bind(self, binding):
        return asinh(bind(self.a, binding))

    def _differentiate(self, variable):
        da = differentiate(self.a, variable)
        return da / sqrt(self.a*self.a + 1.0)

    def __repr__(self):
        return "asinh(%s)"%(repr(self.a))


class ATan(CachedExpression):
    def __init__(self, a):
        super(ATan, self).__init__()
        self.a = a

    def _bind(self, binding):
        return atan(bind(self.a, binding))

    def _differentiate(self, variable):
        da = differentiate(self.a, variable)
        return da / (self.a*self.a + 1.0)

    def __repr__(self):
        return "atan(%s)"%(repr(self.a))


class ATan2(CachedExpression):
    def __init__(self, a, b):
        super(ATan2, self).__init__()
        self.a = a
        self.b = b

    def _bind(self, binding):
        return atan2(bind(self.a, binding), bind(self.b, binding))

    def _differentiate(self, variable):
        a = self.a / self.b
        da = differentiate(a, variable)
        return da / (a*a + 1.0)

    def __repr__(self):
        return "atan2(%s, %s)"%(repr(self.a), repr(self.b))


class ATanh(CachedExpression):
    def __init__(self, a):
        super(ATanh, self).__init__()
        self.a = a

    def _bind(self, binding):
        return atanh(bind(self.a, binding))

    def _differentiate(self, variable):
        da = differentiate(self.a, variable)
        return da / (1.0 - self.a*self.a)

    def __repr__(self):
        return "atanh(%s)"%(repr(self.a))


class Cos(CachedExpression):
    def __init__(self, a):
        super(Cos, self).__init__()
        self.a = a

    def _bind(self, binding):
        return cos(bind(self.a, binding))

    def _differentiate(self, variable):
        da = differentiate(self.a, variable)
        return da * -sin(self.a)

    def __repr__(self):
        return "cos(%s)"%(repr(self.a))


class Cosh(CachedExpression):
    def __init__(self, a):
        super(Cosh, self).__init__()
        self.a = a

    def _bind(self, binding):
        return cosh(bind(self.a, binding))

    def _differentiate(self, variable):
        da = differentiate(self.a, variable)
        return da * sinh(self.a)

    def __repr__(self):
        return "cosh(%s)"%(repr(self.a))


class Sin(CachedExpression):
    def __init__(self, a):
        super(Sin, self).__init__()
        self.a = a

    def _bind(self, binding):
        return sin(bind(self.a, binding))

    def _differentiate(self, variable):
        da = differentiate(self.a, variable)
        return da * cos(self.a)

    def __repr__(self):
        return "sin(%s)"%(repr(self.a))


class Sinh(CachedExpression):
    def __init__(self, a):
        super(Sinh, self).__init__()
        self.a = a

    def _bind(self, binding):
        return sinh(bind(self.a, binding))

    def _differentiate(self, variable):
        da = differentiate(self.a, variable)
        return da * cosh(self.a)

    def __repr__(self):
        return "sinh(%s)"%(repr(self.a))


class Tan(CachedExpression):
    def __init__(self, a):
        super(Tan, self).__init__()
        self.a = a

    def _bind(self, binding):
        return tan(bind(self.a, binding))

    def _differentiate(self, variable):
        da = differentiate(self.a, variable)
        return da / (cos(self.a)*cos(self.a))

    def __repr__(self):
        return "tan(%s)"%(repr(self.a))


class Tanh(CachedExpression):
    def __init__(self, a):
        super(Tanh, self).__init__()
        self.a = a

    def _bind(self, binding):
        return tanh(bind(self.a, binding))

    def _differentiate(self, variable):
        da = differentiate(self.a, variable)
        return da / (cosh(self.a)*cosh(self.a))

    def __repr__(self):
        return "tanh(%s)"%(repr(self.a))


sqrt = wrap_function(SquareRoot, math.sqrt)
log = wrap_function(Logarithm, math.log)

acos = wrap_function(ACos, math.acos)
acosh = wrap_function(ACosh, math.acosh)
asin = wrap_function(ASin, math.asin)
asinh = wrap_function(ASinh, math.asinh)
atan = wrap_function(ATan, math.atan)
atan2 = wrap_function(ATan2, math.atan2)
atanh = wrap_function(ATanh, math.atanh)
cos = wrap_function(Cos, math.cos)
cosh = wrap_function(Cosh, math.cosh)
sin = wrap_function(Sin, math.sin)
sinh = wrap_function(Sinh, math.sinh)
tan = wrap_function(Tan, math.tan)
tanh = wrap_function(Tanh, math.tanh)



from setuptools import setup

setup(name='LDB_Algebra',
      version='0.1.4',
      description='An algebra/calculus package for the LDB framework',
      author='Alex Orange',
      author_email='alex@eldebe.org',
      packages=['ldb', 'ldb.algebra'],
      namespace_packages=['ldb'],
      url='http://www.eldebe.org/ldb/algebra/',
      license='AGPLv3',
      test_suite='test',
     )



# Copyright 2014 Alex Orange
# 
# This file is part of LDB Algebra.
# 
# LDB Algebra is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# LDB Algebra is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with LDB Algebra.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
import math
from ldb.algebra.expression import (CachedExpression, bind, differentiate,
                                    Expression)
from ldb.algebra.math import sqrt

pi = math.pi

def degrees(expression):
    return 180.0/pi*expression

def radians(expression):
    return pi/180.0*expression
